% This is used for 2D experiments

function [y] = locsolve_direct(i, XAX1,Ai,XAX2, XY1, yi, XY2, tol, sol_prev, XBX1,Bi,XBX2, rhs5, locsolve_prec)
% Solver for the 5x5 block structure

global p_state
global p_lagr
global spat_iters

% The matrix is of the form
% M1  0   A'
% 0   M2  N' 
% A   N   0

if (i==1) % Spatial TT block
    % We know that XAX1 contains only ones, since it is the dummy zeroth TT core
    % Ai contains the spatial blocks
    % XAX2 contains the Galerkin reductions of (G0 x D), etc.

    Bs = Bi{1}{1}'; % Bi{1} is of the form [B' 0]
    % Extract the sizes
    [np,nv]=size(Bs);
    np = np/2;
    Bs = Bs(1:np,:);
    r = size(XAX2{3,1}{1},2);
    
    n = [nv; np; nv; nv; np];
    
    pos = cumsum([1; n*r]);
    pos_full = cumsum([1;nv*r*ones(3,1)]);
    
    % What is the RHS?
    % rhs{1} = y{1}, rhs{2}=0, rhs{3}=y{2}, rhs{4}=y{3}, rhs{5}=rhs5
    yc = cell(3,1);
    for j=1:3
        yc{j} = reshape(yi{j}{1}, nv, [])*XY2{j}{1};
        yc{j} = yc{j}.';
    end;
    rhs = [yc{1}(:); zeros(np*r,1); yc{2}(:); yc{3}(:); kron(rhs5, XY2{3}{1}(:))];
    
    % ... and the initial guess
    for j=1:3
        yc{j} = sol_prev(pos_full(j):pos_full(j+1)-1);
        yc{j} = reshape(yc{j}, nv, r);
        yc{j} = yc{j}.';
    end;
    
    % Construct M1 using A{1,1}
    % We know that M1 is rank-1
    M1 = cell(2,1);
    M1{1} = reshape(XAX2{1,1}{1}, r, r);
    M1{2} = Ai{1,1}{1};
    
    % Construct M2 using A{3,3}
    % M2 is rank-1
    M2 = cell(2,1);
    M2{1} = reshape(XAX2{2,2}{1}, r, r);
    M2{2} = Ai{2,2}{1};
    
    % Inverse of diagonal of M2
    iM2 = M2;
    iM2{1} = inv(M2{1});
    iM2{2} = spdiags(1./diag(M2{2}), 0, nv, nv);    
    
    % Construct A
    A = cell(2,3);
    for j=1:3
        A{1,j} = reshape(XAX2{3,1}{j}, r, r);
        A{2,j} = Ai{3,1}{j};
    end;
    Ap = A;
    for j=1:6
        Ap{j} = A{j}';
    end;
    
    % B
    B = cell(2,1);
    B{1} = reshape(XBX2{1}{1}, r, r);
    B{2} = Bs;
    % B'
    Bp = B;
    for j=1:2
        Bp{j} = B{j}';
    end;    
    
    % Construct N
    N = cell(2,1);
    N{1} = reshape(XAX2{3,2}{1}, r, r);
    N{2} = Ai{3,2}{1};
    % N'
    Np = N;
    for j=1:2
        Np{j} = N{j}';
    end;
    
    % Try to reconstruct the initial guess for pressures via LS
    % B'*p_lagr = b1-A*y_lagr-M1*y_state
    % B'*P_state = b3-A*y_state-N*u
    if ((size(p_lagr,2)~=r)||(size(p_state,2)~=r))
        BB = {B{1}; Bs*Bs'};
        rhs_p = rhs(pos(1):pos(2)-1) - canmul(A,yc{3}(:)) - canmul(M1, yc{1}(:));
        rhs_p = reshape(rhs_p, r, nv);
        rhs_p = rhs_p*Bs.';
        p_lagr = rank1div(BB, rhs_p(:));
        rhs_p = rhs(pos(4):pos(5)-1) - canmul(A,yc{1}(:)) - canmul(N, yc{2}(:));
        rhs_p = reshape(rhs_p, r, nv);
        rhs_p = rhs_p*Bs.';
        p_state = rank1div(BB, rhs_p(:));
    else
        p_lagr = p_lagr.';
        p_state = p_state.';
    end;
    
    sol_prev = [yc{1}(:); p_state(:); yc{2}(:); yc{3}(:); p_lagr(:)];
    
    if (locsolve_prec==1)
        % SC1 prec (non-pivoted schur complement)
        % Determine beta
        ind = find(diag(M1{2})>0, 1, 'first');
        beta = M2{2}(ind,ind)/M1{2}(ind,ind);
        
        % Augment M1 to make it invertible
        M1aug = M1;
        zeroind = find(diag(M1aug{2})==0);
        M1aug{2}(zeroind, zeroind) = M2{2}(zeroind,zeroind)*M1{2}(ind,ind); % recreates ~tau*beta*h^2
        M1aug{2} = blkdiag(M1aug{2}, M2{2}(zeroind(1),zeroind(1))*M1{2}(ind,ind)*speye(np));
        
        % Schur complement factors
        Sl = A(:,1);
        Sl{2} = A{2,1} + M1{2}/sqrt(beta)*norm(M1{1})/norm(A{1,1});
        Sl{2} = [Sl{2}, Bp{2}; B{2}, sparse(np,np)];
        
        Sr = Ap(:,1);
        Sr{2} = Ap{2,1} + M1{2}/sqrt(beta)*norm(M1{1})/norm(Ap{1,1});
        Sr{2} = [Sr{2}, Bp{2}; B{2}, sparse(np,np)];
        
        [y,~,~,iters] = minres(@(x)kktmatvec(x,M1,M2,A,Ap,B,Bp,N,Np,pos), rhs, tol, 250, @(x)sc1prec(x,M1aug,M2,Sl,Sr, (nv+np)*r,nv*r,(nv+np)*r), [], sol_prev);
    end;
        
    if (locsolve_prec==2)
        % SC2 prec (pivoted schur complement)
        % Q = N*inv(M2)*N^T matrix
        Q = cell(2,1);
        Q{1} = N{1}*iM2{1}*N{1}';
        Q{1} = (Q{1}+Q{1}')*0.5;
        Q{2} = diag(diag(N{2}))*iM2{2}*N{2}';
        Q{2} = (Q{2}+Q{2}')*0.5;
        % Determine rescaling (to equalize norms)
        scaling = sqrt(norm(M1{1})*norm(M1{2},1)/(norm(Q{1})*norm(Q{2},1)));
        
        % Schur complement factors
        Sl = Ap(:,1);
        Sl{2} = Ap{2,1}+M1{2}/scaling*norm(M1{1})/norm(Ap{1,1});
        Sl{2} = [Sl{2}, Bp{2}; B{2}, sparse(np,np)];
        
        Sr = A(:,1);
        Sr{2} = A{2,1}+Q{2}*scaling*norm(Q{1})/norm(A{1,1});
        Sr{2} = [Sr{2}, Bp{2}; B{2}, sparse(np,np)];
        
        % Expand everything to Stokes systems
        Astokes = A(:,1);
        Astokes{2} = [A{2,1}, Bp{2}; B{2}, sparse(np,np)];
        
        Nstokes = N;
        Nstokes{2} = [N{2}; sparse(np,nv)];
        Npstokes = Np;
        Npstokes{2} = [Np{2}, sparse(nv,np)];
                
        [y,iters] = fgmres(@(x,tol)kktmatvec(x,M1,M2,A,Ap,B,Bp,N,Np,pos), rhs, tol, 'P', @(x,tol)sc2prec(x,M2,Astokes,Sl,Sr,Nstokes,Npstokes, (nv+np)*r,nv*r,(nv+np)*r), 'x0', sol_prev, 'restart', 250, 'max_iters', 1);
        iters = iters(2);
    end;
    
    spat_iters = spat_iters+iters;
    
    % Permute the solution back to r1 x n x r2 order
    yc = cell(5,1);
    for j=1:5
        yc{j} = y(pos(j):pos(j+1)-1);
        yc{j} = reshape(yc{j}, r, n(j));
        yc{j} = yc{j}.';
    end;
    % Now we only need y{1},y{3},y{4}
    y = [yc{1}(:); yc{3}(:); yc{4}(:)];
    % Store pressures for the final assembly
    p_state = yc{2}; % size n,r
    p_lagr = yc{5};
    
    % Correction to the RHS
    df = [p_lagr; p_state];
    y = {y, df};
    
else % Time or stochastic TT blocks
    % SC1 is not defined, work with SC2 only.

    % Ranks
    r1 = size(XAX1{3,1}{1}, 1);
    r2 = size(XAX2{3,1}{1}, 2);
    n = size(Ai{3,1}{1},1);
    
    nf = r1*n*r2;

    pos = cumsum([1;nf*ones(3,1)]);
    
    % Create the RHS
    yc = cell(3,1);
    for j=1:3
        yc{j} = zeros(r2, r1*n);
        for m=1:size(yi{j},2)
            [ry1,~,ry2]=size(yi{j}{m});
            ym = reshape(yi{j}{m}, ry1,n*ry2);
            ym = XY1{j}{m}*ym;
            ym = reshape(ym, r1*n, ry2);
            ym = ym*XY2{j}{m};
            ym = reshape(ym, r1*n, r2);
            ym = ym.';
            yc{j} = yc{j}+ym;
        end;
    end;
    rhs = [yc{1}(:); yc{2}(:); yc{3}(:)];
    
    % Permute the previous solution
    for j=1:3
        yc{j} = sol_prev(pos(j):pos(j+1)-1);
        yc{j} = reshape(yc{j}, r1*n, r2);
        yc{j} = yc{j}.';
    end;
    sol_prev = [yc{1}(:); yc{2}(:); yc{3}(:)];
    
    % A (only the velocity part)
    A = cell(3,3);
    for j=1:3
        A{1,j} = reshape(XAX2{3,1}{j}, r2, r2);
        A{2,j} = reshape(XAX1{3,1}{j}, r1, r1);
        A{3,j} = Ai{3,1}{j};
    end;
    Ap = A;
    for j=1:9
        Ap{j} = A{j}';
    end;
   
    % M1
    M1 = cell(3,1);
    M1{1} = reshape(XAX2{1,1}{1}, r2, r2);  M1{1} = (M1{1}+M1{1}')*0.5;
    M1{2} = reshape(XAX1{1,1}{1}, r1, r1);  M1{2} = (M1{2}+M1{2}')*0.5;
    M1{3} = Ai{1,1}{1};
        
    % M2
    M2 = cell(3,1);
    M2{1} = reshape(XAX2{2,2}{1}, r2, r2);  M2{1} = (M2{1}+M2{1}')*0.5;
    M2{2} = reshape(XAX1{2,2}{1}, r1, r1);  M2{2} = (M2{2}+M2{2}')*0.5;
    M2{3} = Ai{2,2}{1};    

    % N (only for velocity)
    N = cell(3,1);
    N{1} = reshape(XAX2{3,2}{1}, r2, r2);
    N{2} = reshape(XAX1{3,2}{1}, r1, r1);
    N{3} = Ai{3,2}{1};

    % N'
    Np = N;
    for j=1:3
        Np{j} = N{j}';
    end;
    
    % Q = N*inv(M2)*N'
    Q = cell(3,1);
    Q{1} = squeeze(XAX2{3,2}{1})/squeeze(XAX2{2,2}{1})*squeeze(XAX2{3,2}{1})';
    Q{1} = (Q{1}+Q{1}')*0.5;
    Q{2} = squeeze(XAX1{3,2}{1})/squeeze(XAX1{2,2}{1})*squeeze(XAX1{3,2}{1})';
    Q{2} = (Q{2}+Q{2}')*0.5;
    Q{3} = Ai{3,2}{1}/Ai{2,2}{1}*Ai{3,2}{1}';
    Q{3} = (Q{3}+Q{3}')*0.5;
    
    % Rescaling for the Schur complement factors
    scaling = sqrt((norm(M1{1},1)*norm(M1{2},1)*norm(M1{3},1))/(norm(Q{1},1)*norm(Q{2},1)*norm(Q{3},1)));

    if (i==2)
        % The total size is still too large, given the matrices are dense.
        % Use the rank-1 preconditioner as well
        Sl = Ap(:,1);
        Sl{3} = Ap{3,1}+M1{3}/scaling*norm(M1{2},1)*norm(M1{1},1)/norm(Ap{2,1},1)/norm(Ap{1,1},1);
        Sr = A(:,1);
        Sr{3} = A{3,1}+Q{3}*scaling*norm(Q{2},1)*norm(Q{1},1)/norm(A{2,1},1)/norm(A{1,1},1);    
        
        Aprec = A(:,1);
    else % i==3, for the time block we can expand the kron products
        Aprec = sparse(nf,nf);
        for j=1:3
            Aprec = Aprec+kron(A{3,j}, kron(sparse(A{2,j}), sparse(A{1,j})));
        end;
        
        M1full = kron(M1{3}, kron(sparse(M1{2}), sparse(M1{1})));
        Qfull = kron(Q{3}, kron(sparse(Q{2}), sparse(Q{1})));
        
        Sl = Aprec'+M1full/scaling;
        Sr = Aprec+Qfull*scaling;
    end;
    
    % SC2 prec
    y = fgmres(@(x,tol)kktmatvec3(x,M1,M2,A,Ap,N,Np,nf,nf,nf), rhs, tol, 'P', @(x,tol)sc2prec(x,M2,Aprec,Sl,Sr,N,Np, nf,nf,nf), 'x0', sol_prev, 'restart', 250, 'max_iters', 1);
        
    yc = cell(3,1);
    for j=1:3
        yc{j} = y(pos(j):pos(j+1)-1);
        yc{j} = reshape(yc{j}, r2, r1*n);
        yc{j} = yc{j}.';
    end;
    y = [yc{1}(:); yc{2}(:); yc{3}(:)];
end;

end



function [y]=kktmatvec(x,M1,M2,A,Ap,B,Bp,N,Np,pos)
% For 5x5 structure in space
x1 = x(pos(1):pos(2)-1);
x2 = x(pos(2):pos(3)-1);
x3 = x(pos(3):pos(4)-1);
x4 = x(pos(4):pos(5)-1);
x5 = x(pos(5):pos(6)-1);

y1 = canmul(M1,x1)+canmul(Ap,x4)+canmul(Bp,x5);
y2 = canmul(B,x4);
y3 = canmul(M2,x3)+canmul(Np,x4);
y4 = canmul(A,x1)+canmul(Bp,x2)+canmul(N,x3);
y5 = canmul(B,x1);
y = [y1;y2;y3;y4;y5];
end

function [y]=kktmatvec3(x,M1,M2,A,Ap,N,Np,n1,n2,n3)
% For 3x3 structure in time, since the pressure is excluded
x1 = x(1:n1);
x2 = x(n1+1:n1+n2);
x3 = x(n1+n2+1:n1+n2+n3);
y1 = canmul(M1,x1)+canmul(Ap,x3);
y2 = canmul(M2,x2)+canmul(Np,x3);
y3 = canmul(A,x1)+canmul(N,x2);
y = [y1;y2;y3];
end

function [y]=canmul(A,x)
% Multiplies the canonical dD matrix by a full vector
[d,R] = size(A);
n = zeros(1,d);
m = zeros(1,d);
for i=1:d
    [n(i),m(i)]=size(A{i,1});
end;
y = zeros(prod(n),1);
for j=1:R
    yj = x;
    for i=1:d
        yj = reshape(yj, m(i), []);
        yj = A{i,j}*yj;
        yj = yj.';
    end;
    yj = yj(:);
    y = y+yj;
end;
end


function [y]=rank1div(A,x)
% Solves the system with the kron prod of d matrices
d = size(A,1);
n = zeros(1,d);
for i=1:d
    n(i)=size(A{i,1},1);
end;
y = x;
for i=1:d
    y = reshape(y, n(i), []);
    y = A{i,1}\y;
    y = y.';
end;
y = y(:);
end



function [y]=sc2prec(x,M2,A,Sl,Sr,N,Np, n1,n2,n3)
% Triangular SC2 preconditioner
if (isa(Sl, 'cell'))
    % Only rank-1 operations for space and stochastics
    y3 = rank1div(Sl,x(1:n1));
    y3 = canmul(A, y3);
    y3 = rank1div(Sr,y3);
    
    y21 = canmul(Np,y3);
    y2 = x(n1+1:n1+n2) - y21;
    y2 = rank1div(M2,y2);
    
    y12 = canmul(N,y2);
    y1 = x(n1+n2+1:n1+n2+n3) - y12;
    y1 = rank1div(A,y1);
else
    % Full backslash -- for time
    y3 = Sl\x(1:n1);
    y3 = A*y3;
    y3 = Sr\y3;
    
    y21 = canmul(Np,y3);
    y2 = x(n1+1:n1+n2) - y21;
    y2 = rank1div(M2,y2);
    
    y12 = canmul(N,y2);
    y1 = x(n1+n2+1:n1+n2+n3) - y12;
    y1 = A\y1;
end;

y = [y1;y2;y3];
end



% function [y]=sc2prec_cell(x,M2,A,Sl,Sr,N,Np, n1,n2,n3)
% y3 = rank1div(Sl, x(1:n1));
% y3 = canmul(A,y3);
% y3 = rank1div(Sr, y3);
% 
% y21 = canmul(Np,y3);
% y2 = x(n1+1:n1+n2) - y21;
% y2 = rank1div(M2,y2);
% 
% y12 = canmul(N,y2);
% y1 = x(n1+n2+1:n1+n2+n3) - y12;
% y1 = rank1div(A, y1);
% 
% y = [y1;y2;y3];
% end


function [y]=sc1prec(x,M1,M2,Sl,Sr, n1, n2, n3)
% Diagonal SC1 prec
y1 = rank1div(M1, x(1:n1));
y2 = rank1div(M2, x(n1+1:n1+n2));

y3 = rank1div(Sl, x(n1+n2+1:n1+n2+n3));
y3 = canmul(M1, y3);
y3 = rank1div(Sr, y3);

y = [y1;y2;y3];
end
