% Theis is used for 3D experiments

function [y] = locsolve_iter(i, XAX1,Ai,XAX2, XY1, yi, XY2, tol, sol_prev, XBX1,Bi,XBX2, rhs5, P_stokes)
% Solver for the 5x5 block structure

global p_state
global p_lagr
global spat_iters

% The matrix should be of the form
% M1  A'  B'
% 0   M2  N' 
% A   N   0

if (i==1) % Spatial block
    % We know that XAX1 contains only ones, since it is the dummy zeroth TT core
    % Ai contains the spatial blocks
    % XAX2 contains the Galerkin reductions of (G0 x D), etc.

    Bs = Bi{1}{1}'; % Bi{1} is of the form [B' 0]
    % Extract the sizes
    [np,nv]=size(Bs);
    np = np/2;
    Bs = Bs(1:np,:);
    r = size(XAX2{3,1}{1},2);
    
    n = [nv; np; nv; nv; np];
    
    pos = cumsum([1; n*r]);
    pos_full = cumsum([1;nv*r*ones(3,1)]);
    
    % What is the RHS?
    % rhs{1} = y{1}, rhs{2}=0, rhs{3}=y{2}, rhs{4}=y{3}, rhs{5}=rhs5    
    yc = cell(3,1);
    yc{1} = reshape(yi{1}{1}, nv, [])*XY2{1}{1};
    yc{2} = reshape(yi{2}{1}, nv, [])*XY2{2}{1};
    yc{3} = reshape(yi{3}{1}, nv, [])*XY2{3}{1}; 
    rhs = [yc{1}(:); zeros(np*r,1); yc{2}(:); yc{3}(:); reshape(rhs5*XY2{3}{1}, np*r, 1)];
    
    % ... and the initial guess
    for j=1:3
        yc{j} = sol_prev(pos_full(j):pos_full(j+1)-1);
        yc{j} = reshape(yc{j}, nv, r);
    end;
    
    % Construct M1 using A{1,1}
    % We know that M1 is rank-1
    M1 = {Ai{1,1}{1}; squeeze(XAX2{1,1}{1})};
    
    % Construct M2 using A{3,3}
    % M2 is rank-1
    M2 = {Ai{2,2}{1}; squeeze(XAX2{2,2}{1})};    
        
    % Construct A
    A_space = {Ai{3,1}{1}, Ai{3,1}{2}, Ai{3,1}{3}};
    A_time = {squeeze(XAX2{3,1}{1}), squeeze(XAX2{3,1}{2}), squeeze(XAX2{3,1}{3})};
    Ap_time = {squeeze(XAX2{3,1}{1})', squeeze(XAX2{3,1}{2})', squeeze(XAX2{3,1}{3})'};
    
    A = [A_space; A_time];
    Ap = [A_space; Ap_time];
    
    % B
    B = {Bs; reshape(XBX2{1}{1}, r, r)};
    Bp = {Bs'; reshape(XBX2{1}{1}, r, r)'};
    
    % Construct N
    N = {Ai{3,2}{1}; squeeze(XAX2{3,2}{1})};
    Np = {Ai{3,2}{1}'; squeeze(XAX2{3,2}{1})'};    
    
    % Try to reconstruct the initial guess for pressures via LS
    % B'*p_lagr = b1-A'*y_lagr-M1*y_state
    % B'*P_state = b3-A*y_state-N*u
    if ((size(p_lagr,2)~=r)||(size(p_state,2)~=r))
        BB = Bs*Bs';
        rhs_p = reshape(yi{1}{1}, nv, [])*XY2{1}{1};
        rhs_p = rhs_p(:) - canmul(Ap,yc{3}) - canmul(M1, yc{1});
        rhs_p = reshape(rhs_p, nv, r);
        rhs_p = Bs*rhs_p;
        p_lagr = BB\rhs_p;
        p_lagr = p_lagr/B{2};
        rhs_p = reshape(yi{3}{1}, nv, [])*XY2{3}{1};
        rhs_p = rhs_p(:) - canmul(A,yc{1}) - canmul(N, yc{2});
        rhs_p = reshape(rhs_p, nv, r);
        rhs_p = Bs*rhs_p;
        p_state = BB\rhs_p;
        p_state = p_state/B{2};
    end;
    
    sol_prev = [yc{1}(:); p_state(:); yc{2}(:); yc{3}(:); p_lagr(:)];

    ind = find(diag(M1{1})>0, 1, 'first');
    beta = M2{1}(ind,ind)/M1{1}(ind,ind);
    
    % SC2 prec
    Q = cell(2,1);
    Q{1} = M1{1}/beta;
    Q{2} = N{2}/M2{2}*Np{2};
    Q{2} = (Q{2}+Q{2}')*0.5;
    scaling = sqrt(beta)*sqrt(norm(M1{2})/norm(Q{2}));

    % Left part
    Sl = [Ap, M1];
    Sl{1,4} = M1{1}/scaling;    
    % Pressure operator
    Slp = Sl;
    Slp(1,1:3) = P_stokes.Fp;
    Slp{1,4} = P_stokes.Mp/scaling;

    % Right part
    Sr = [A, Q];
    Sr{1,4} = Q{1}*scaling;
    % Pressure operator
    Srp = Sr;
    Srp(1,1:3) = P_stokes.Fp;
    Srp{1,4} = P_stokes.Mp/scaling;    
    

    % Prepare the diagonal-in-time matrices and coarsen them
    control = hsl_mi20_control;
    A_d = cell(r,1);
    Sl_d = cell(r,1);
    Sr_d = cell(r,1);
    Ra = [];
    for j=1:r
        A_d{j} = A{1,1}*A{2,1}(j,j) + A{1,2}*A{2,2}(j,j) + A{1,3}*A{2,3}(j,j);
        Sl_d{j} = Sl{1,1}*Sl{2,1}(j,j) + Sl{1,2}*Sl{2,2}(j,j) + Sl{1,3}*Sl{2,3}(j,j) + Sl{1,4}*Sl{2,4}(j,j);
        Sr_d{j} = Sr{1,1}*Sr{2,1}(j,j) + Sr{1,2}*Sr{2,2}(j,j) + Sr{1,3}*Sr{2,3}(j,j) + Sr{1,4}*Sr{2,4}(j,j);
        
        if (isempty(Ra))
            try
                [info,cd1,cd2]=hsl_mi20_setup_cdata(A_d{j}, control);
                Ra = extract_restrictor(info,cd1,cd2);
                clear cd1 cd2
            catch me
                fprintf('Coarsening of A at j=%d failed with message:\n%s\n', j, me.message);
            end;
        end;
        if (isempty(Ra))
            try
                [info,cd1,cd2]=hsl_mi20_setup_cdata(Sl_d{j}, control);
                Ra = extract_restrictor(info,cd1,cd2);
                clear cd1 cd2
            catch me
                fprintf('Coarsening of Sl at j=%d failed with message:\n%s\n', j, me.message);
            end;
        end;
        if (isempty(Ra))
            try
                [info,cd1,cd2]=hsl_mi20_setup_cdata(Sr_d{j}, control);
                Ra = extract_restrictor(info,cd1,cd2);
                clear cd1 cd2
            catch me
                fprintf('Coarsening of Sr at j=%d failed with message:\n%s\n', j, me.message);
            end;
        end;        
        
        if (~isempty(Ra))
            A_d{j,1} = mg_coarsen(A_d{j}, Ra);
            Sl_d{j,1} = mg_coarsen(Sl_d{j}, Ra);
            Sr_d{j,1} = mg_coarsen(Sr_d{j}, Ra);
        end;
    end;
    if (isempty(Ra))
        error('All coarsenings failed');
    end;
    
    % Coarsen M2
    % I have no idea why MI20 fails on coarsening of Mv2... But it
    % seems that the restrictors from A work okay
    M2c = mg_coarsen(M2{1}, Ra);
    
    % More pressure ops
    Fp = [P_stokes.Fp; A_time];
    Mp = {P_stokes.Mp; A_time{1}};
    Kp = {P_stokes.Kp; A_time{1}};
    
    [y,iters] = fgmres(@(x,tol)kktmatvec(x,M1,M2,A,Ap,B,Bp,N,Np,pos), rhs, tol, 'P', @(x,tol)sc2prec_cell(x,A,A_d,Ra,Fp,Sl,Sl_d,Slp,Sr,Sr_d,Srp, M2,M2c, N,Np, B,Bp, Kp, Mp, nv,np,r), 'x0', sol_prev, 'restart', 250, 'max_iters', 1);
    
    iters = iters(2);
    spat_iters = spat_iters+iters;
    
    % Permute the solution back to r1 x n x r2 order
    yc = cell(5,1);
    for j=1:5
        yc{j} = y(pos(j):pos(j+1)-1);
    end;
    % Now we only need y{1},y{3},y{4}
    y = [yc{1}(:); yc{3}(:); yc{4}(:)];
    % Store pressures for the final assembly
    p_state = reshape(yc{2}, np, r); % size n,r
    p_lagr = reshape(yc{5}, np, r);
    
    % Correction to the RHS
    df = [p_lagr; p_state];
    y = {y, df};  
    
else % time, stochastic blocks
    % i>1 -- matrices are relatively small.
    % SC2 prec should work

    % Ranks
    r1 = size(XAX1{3,1}{1}, 1);
    r2 = size(XAX2{3,1}{1}, 2);
    n = size(Ai{3,1}{1},1);
    
    nf = r1*n*r2;

    pos = cumsum([1;nf*ones(3,1)]);
    
    % Create the RHS
    yc = cell(3,1);
    for j=1:3
        yc{j} = zeros(r2, r1*n);
        for m=1:size(yi{j},2)
            [ry1,~,ry2]=size(yi{j}{m});
            ym = reshape(yi{j}{m}, ry1,n*ry2);
            ym = XY1{j}{m}*ym;
            ym = reshape(ym, r1*n, ry2);
            ym = ym*XY2{j}{m};
            ym = reshape(ym, r1*n, r2);
            ym = ym.';
            yc{j} = yc{j}+ym;
        end;
    end;
    rhs = [yc{1}(:); yc{2}(:); yc{3}(:)];
    
    % Permute the previous solution
    for j=1:3
        yc{j} = sol_prev(pos(j):pos(j+1)-1);
        yc{j} = reshape(yc{j}, r1*n, r2);
        yc{j} = yc{j}.';
    end;
    sol_prev = [yc{1}(:); yc{2}(:); yc{3}(:)];
    
    % A (only the velocity part)
    A = cell(3,3);
    for j=1:3
        A{1,j} = reshape(XAX2{3,1}{j}, r2, r2);
        A{2,j} = reshape(XAX1{3,1}{j}, r1, r1);
        A{3,j} = Ai{3,1}{j};
    end;
    Ap = A;
    for j=1:9
        Ap{j} = A{j}';
    end;
   
    % M1
    M1 = cell(3,1);
    M1{1} = reshape(XAX2{1,1}{1}, r2, r2);  M1{1} = (M1{1}+M1{1}')*0.5;
    M1{2} = reshape(XAX1{1,1}{1}, r1, r1);  M1{2} = (M1{2}+M1{2}')*0.5;
    M1{3} = Ai{1,1}{1};
        
    % M2
    M2 = cell(3,1);
    M2{1} = reshape(XAX2{2,2}{1}, r2, r2);  M2{1} = (M2{1}+M2{1}')*0.5;
    M2{2} = reshape(XAX1{2,2}{1}, r1, r1);  M2{2} = (M2{2}+M2{2}')*0.5;
    M2{3} = Ai{2,2}{1};    
        
    % N (only for velocity)
    N = cell(3,1);
    N{1} = reshape(XAX2{3,2}{1}, r2, r2);
    N{2} = reshape(XAX1{3,2}{1}, r1, r1);
    N{3} = Ai{3,2}{1};

    % N'
    Np = N;
    for j=1:3
        Np{j} = N{j}';
    end;
    
    % Q = N*inv(M2)*N'
    Q = cell(3,1);
    Q{1} = squeeze(XAX2{3,2}{1})/squeeze(XAX2{2,2}{1})*squeeze(XAX2{3,2}{1})';
    Q{1} = (Q{1}+Q{1}')*0.5;
    Q{2} = squeeze(XAX1{3,2}{1})/squeeze(XAX1{2,2}{1})*squeeze(XAX1{3,2}{1})';
    Q{2} = (Q{2}+Q{2}')*0.5;
    Q{3} = Ai{3,2}{1}/Ai{2,2}{1}*Ai{3,2}{1}';
    Q{3} = (Q{3}+Q{3}')*0.5;
    
    % Rescaling for Schur Prec factors
    scaling = sqrt((norm(M1{1},1)*norm(M1{2},1)*norm(M1{3},1))/(norm(Q{1},1)*norm(Q{2},1)*norm(Q{3},1)));

    if (i==2)
        % use the rank-1 form, keeping the full matrices is too expensive
        Sl = Ap(:,1);
        Sl{3} = Ap{3,1}+M1{3}/scaling*norm(M1{2},1)*norm(M1{1},1)/norm(Ap{2,1},1)/norm(Ap{1,1},1);
        Sr = A(:,1);
        Sr{3} = A{3,1}+Q{3}*scaling*norm(Q{2},1)*norm(Q{1},1)/norm(A{2,1},1)/norm(A{1,1},1);    
        
        Aprec = A(:,1);
    else % i==3, for time block we can expand the kron products
        Aprec = sparse(nf,nf);
        for j=1:3
            Aprec = Aprec+kron(A{3,j}, kron(sparse(A{2,j}), sparse(A{1,j})));
        end;
        
        M1full = kron(M1{3}, kron(sparse(M1{2}), sparse(M1{1})));
        Qfull = kron(Q{3}, kron(sparse(Q{2}), sparse(Q{1})));
        
        Sl = Aprec'+M1full/scaling;
        Sr = Aprec+Qfull*scaling;
    end;
    
    % SC2 prec
    y = fgmres(@(x,tol)kktmatvec3(x,M1,M2,A,Ap,N,Np,nf,nf,nf), rhs, tol, 'P', @(x,tol)sc2prec(x,M2,Aprec,Sl,Sr,N,Np, nf,nf,nf), 'x0', sol_prev, 'restart', 300, 'max_iters', 1);
        
    yc = cell(3,1);
    for j=1:3
        yc{j} = y(pos(j):pos(j+1)-1);
        yc{j} = reshape(yc{j}, r2, r1*n);
        yc{j} = yc{j}.';
    end;
    y = [yc{1}(:); yc{2}(:); yc{3}(:)];
end;

end



function [y]=kktmatvec(x,M1,M2,A,Ap,B,Bp,N,Np,pos)
x1 = x(pos(1):pos(2)-1);
x2 = x(pos(2):pos(3)-1);
x3 = x(pos(3):pos(4)-1);
x4 = x(pos(4):pos(5)-1);
x5 = x(pos(5):pos(6)-1);

y1 = canmul(M1,x1)+canmul(Ap,x4)+canmul(Bp,x5);
y2 = canmul(B,x4);
y3 = canmul(M2,x3)+canmul(Np,x4);
y4 = canmul(A,x1)+canmul(Bp,x2)+canmul(N,x3);
y5 = canmul(B,x1);
y = [y1;y2;y3;y4;y5];
end


function [y]=kktmatvec3(x,M1,M2,A,Ap,N,Np,n1,n2,n3)
x1 = x(1:n1);
x2 = x(n1+1:n1+n2);
x3 = x(n1+n2+1:n1+n2+n3);
y1 = canmul(M1,x1)+canmul(Ap,x3);
y2 = canmul(M2,x2)+canmul(Np,x3);
y3 = canmul(A,x1)+canmul(N,x2);
y = [y1;y2;y3];
end

function [y]=canmul(A,x)
% Multiplies the canonical dD matrix by a full vector
[d,R] = size(A);
n = zeros(1,d);
m = zeros(1,d);
for i=1:d
    [n(i),m(i)]=size(A{i,1});
end;
y = zeros(prod(n),1);
for j=1:R
    yj = x;
    for i=1:d
        yj = reshape(yj, m(i), []);
        yj = A{i,j}*yj;
        yj = yj.';
    end;
    yj = yj(:);
    y = y+yj;
end;
end


function [y]=rank1div(A,x)
% Solves the system with the kron prod of d matrices
d = size(A,1);
n = zeros(1,d);
for i=1:d
    n(i)=size(A{i,1},1);
end;
y = x;
for i=1:d
    y = reshape(y, n(i), []);
    y = A{i,1}\y;
    y = y.';
end;
y = y(:);
end



function [y]=sc2prec(x,M2,A,Sl,Sr,N,Np, n1,n2,n3)
if (isa(Sl, 'cell'))
    % Only rank-1 operations
    y3 = rank1div(Sl,x(1:n1));
    y3 = canmul(A, y3);
    y3 = rank1div(Sr,y3);
    
    y21 = canmul(Np,y3);
    y2 = x(n1+1:n1+n2) - y21;
    y2 = rank1div(M2,y2);
    
    y12 = canmul(N,y2);
    y1 = x(n1+n2+1:n1+n2+n3) - y12;
    y1 = rank1div(A,y1);
else
    % Full backslash -- for time
    y3 = Sl\x(1:n1);
    y3 = A*y3;
    y3 = Sr\y3;
    
    y21 = canmul(Np,y3);
    y2 = x(n1+1:n1+n2) - y21;
    y2 = rank1div(M2,y2);
    
    y12 = canmul(N,y2);
    y1 = x(n1+n2+1:n1+n2+n3) - y12;
    y1 = A\y1;
end;

y = [y1;y2;y3];
end


function [y]=sc2prec_cell(x,A,A_d,Ra,Ap,Sl,Sl_d,Slp,Sr,Sr_d,Srp, M2,M2c, N,Np, B,Bp, Kp, Mp, nv,np,r)
% The preconditioner with iterative inner solves
y3 = fgmres(@(x,tol)stokesmatvec(Sl,B,Bp,x,nv,np,r), x(1:(nv+np)*r), 1e-5, 'P', @(x,tol)stokes_prec(x,Sl,Sl_d,Ra,Bp, Kp,Slp,Mp), 'restart', 50, 'max_iters', 1, 'verb', 1);
y3 = stokesmatvec(A,B,Bp,y3,nv,np,r);
y3 = fgmres(@(x,tol)stokesmatvec(Sr,B,Bp,x,nv,np,r), y3, 1e-5, 'P', @(x,tol)stokes_prec(x,Sr,Sr_d,Ra,Bp, Kp,Srp,Mp), 'restart', 50, 'max_iters', 1, 'verb', 1);

y21 = stokesmatvec(Np,[],[],y3,nv,np,r);
y21 = y21(1:nv*r);
y2 = x((nv+np)*r+1:(nv+np+nv)*r) - y21;

y2 = pcg(@(x)canmul(M2,x), y2, 1e-8, 10, @(x)jacobiprec(x,M2c,M2{2},Ra));

y12 = stokesmatvec(N,[],[],y2,nv,np,r);
y1 = x((nv+np+nv)*r+1:(nv+np+nv+nv+np)*r) - y12;
y1 = fgmres(@(x,tol)stokesmatvec(A,B,Bp,x,nv,np,r), y1, 1e-5, 'P', @(x,tol)stokes_prec(x,A,A_d,Ra,Bp, Kp,Ap,Mp), 'restart', 50, 'max_iters', 1, 'verb', 1);

y = [y1;y2;y3];
end

function [y]=stokesmatvec(A11,B,Bp,x,nv,np,r)
% MatVec with 2x2 Stokes matrix
y1 = canmul(A11, x(1:nv*r));
if (~isempty(B))
    y1 = y1 + canmul(Bp, x(nv*r+1:(nv+np)*r));
    y2 = canmul(B, x(1:nv*r));
else
    y2 = zeros(np*r,1);
end;
y = [y1;y2];
end

function [y]=jacobiprec(x,A_d,A_time,Ra)
% Block Jacobi multigrid
n = size(A_d{1},1);
r = size(A_time,1);
y = reshape(x, n, r);
y = y/A_time;
for j=1:r
    y(:,j) = mg_cycle_simple(y(:,j), A_d, Ra);
end;
y = y(:);
end

function [y]=stokes_prec(x,A11,A11_d,Ra,Bp, Kp,Fp,Mp)
% Commutator preconditioner for Stokes, with multigrid for velocity, Jacobi
% over time
[nv,np]=size(Bp{1});
r = size(A11{2,1},1);

% Pressure
y2 = x(nv*r+1:(nv+np)*r);

y2 = rank1div(Mp, y2);
y2 = canmul(Fp, y2);
y2 = rank1div(Kp, y2);

y1 = x(1:nv*r);
y1 = y1+canmul(Bp, y2);

% triangular elimination
y1 = reshape(y1, nv, []);
R = size(A11, 2);
for j=1:r
    y1(:,j) = mg_cycle_simple(y1(:,j), A11_d{j}, Ra);
    for k=1:R
        y1(:,j+1:r) = y1(:,j+1:r) - (A11{1,k}*y1(:,j))*A11{2,k}(j+1:r,j).';
    end;
end;

y = [y1(:); -y2(:)];

end
