% The runfile for computing eigenvalues of the preconditioned matrix.
% Note: even for small grids, the matrices are very large after the kron
% products. Use only if you have enough memory on your PC.
% Use gen_SBmat_small.py to generate matrices


N = 2;
 T = 1;
nt = 2^N;
dt = T/nt; 
% Regularization
% beta = 1e-4;
% alpha = 1;
% Tolerance 
tol = 1e-6;
%Stochastics
nu1 = 1e-1;

% Time matrices
It = speye(nt);
Ct = spdiags(-ones(nt,1), -1, nt, nt);

% Stochastic matrices
load G0_G1_small.mat
G0 = sparse(G0);
G1=  sparse(G1);
G1(abs(G1)<1e-15)=0;
G1 = sparse(G1);

D=speye(nt);
% D(1,1)=1/2; The first time point is dt, not 0! Hence the quadrature weight 1
D(nt,nt)=1/2;

% Spatial matrices from Fenics
A=mmread('A.mtx'); % Stokes+Brinkmann
A2=mmread('A2.mtx'); % Stokes only, without boundary nodes -- for stochastic term
B=mmread('C.mtx'); % Funny

Mv=mmread('M.mtx'); % mass without boundary nodes
Mv2=mmread('M2.mtx'); % with boundary nodes

% Pressure matrices
Mp=mmread('Mp.mtx');
Kp=mmread('Kp.mtx');
Kpk=mmread('Kpk.mtx'); % inner(grad(p),grad(q))+k*inner(p,q)

n=size(Mv,2);
P = size(G0,1);

% A11 = {dt*Mv; G_alpha; D};
% 
% % The control mass
% A22 = {dt*beta*Mv2; G0; D};
% 
% % The N-matrix
% A32 = {dt*Mv; G0; It};
% 
% % Now the system matrices
% A31 = {dt*(Mv/dt+A), dt*(nu1*A2), dt*(Mv/dt);
%        G0,          G1,     G0;
%        It,          It,     Ct};   
% A13 = {dt*(Mv/dt+A'),dt*(nu1*A2), dt*(Mv/dt);
%        G0,          G1,     G0;
%        It,          It,     Ct'};   
   
A  = full(A);
Mv = full(Mv);
G0 = full(G0);
G1 = full(G1);
It = full(It);
Ct = full(Ct);
A2 = full(A2);
B = full(B);
%G_alpha = full(G_alpha);
D = full(D);

fprintf('Everything loaded\n');

AA = kron(It, kron(G0,dt*(Mv/dt+A))) + kron(It, kron(G1,dt*(nu1*A2))) + kron(Ct, kron(G0,dt*(Mv/dt)));
BB = kron(It, kron(G0,B));

fprintf('AA, BB done\n');

na =size(AA,2);
nb =size(BB,1);
Iaa = eye(na);
Ibb = eye(nb);

iAAp = (AA')\Iaa;

S0= BB*iAAp*BB';
S0 = S0\Ibb;
Pk= BB'*S0*BB*iAAp;
K11= iAAp*( Iaa - Pk );

fprintf('S0, K11 done\n');

na=3;
nb=5;



%alpha
alpha=zeros(na,1);
alpha(1)=1e-4;
for j=1:na-1
    alpha(j+1)=1e+4*alpha(j);
end


%beta
beta=zeros(nb,1);
beta(1)=1e-8;
for j=1:nb-1
    beta(j+1)=1e+4*beta(j);
end




%For alpha
Ga=zeros(na,n*P*nt);
T1a = G0; 
T1a(1,1)= 0;

%for Beta
Gb=zeros(nb,n*P*nt);
T1b = G0; 
T1b(1,1)= 0;

fprintf('Simulate for a fixed beta\n');

for i=1:na
    tic
% Now  run through each alpha
T2 = alpha(i)*T1a;
G_alpha = full(G0 + T2);

% Now fix beta=1
Mm1 = kron(inv(D), kron(G0,Mv));
M1 = kron(D, kron(G_alpha,Mv));

Astar = AA + (1/beta(3))*Mm1*K11*M1;
iAstar = Astar\Iaa;
Sstar = BB*iAstar*BB';

ns =size(Sstar,2);
Iss = eye(ns);
npk =size(Pk,2);
Ipk = eye(npk);

iSstar = Sstar\Iss;

Pstar = BB'*iSstar*BB*iAstar;
npstar =size(Pstar,2);
Ipstar = eye(npstar);

U =sqrt(1/beta(3))*( Mm1 + (AA*iAAp*(Ipk - Pk) + Pk)*M1);

Pnew = iAstar*(Ipstar - Pstar)*U;

npnew =size(Pnew,2);
Ipnew = eye(npnew);

E = Ipnew + Pnew;

Ga(i,:)=eig(E);
toc
end


fprintf('Simulate for a fixed alpha\n');
for i=1:nb
    tic
% Now fix alpha=1
T2 = alpha(2)*T1b;
G_alpha = full(G0 + T2);

% Now  run through each beta
Mm1 = kron(inv(D), kron(G0,Mv));
M1 = kron(D, kron(G_alpha,Mv));

Astar = AA + (1/beta(i))*Mm1*K11*M1;
iAstar = Astar\Iaa;
Sstar = BB*iAstar*BB';

ns =size(Sstar,2);
Iss = eye(ns);
npk =size(Pk,2);
Ipk = eye(npk);

iSstar = Sstar\Iss;

Pstar = BB'*iSstar*BB*iAstar;
npstar =size(Pstar,2);
Ipstar = eye(npstar);

U =sqrt(1/beta(i))*( Mm1 + (AA*iAAp*(Ipk - Pk) + Pk)*M1);

Pnew = iAstar*(Ipstar - Pstar)*U;

npnew =size(Pnew,2);
Ipnew = eye(npnew);

E = Ipnew + Pnew;

Gb(i,:)=eig(E);
toc
end

save('eigenv.mat', 'Ga', 'Gb');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Plot for a fixed beta')
hold on
x=1:n*P*nt;
nx =length(x);

plot(x,abs(Ga(1,:)),'--ro', x,abs(Ga(2,:)),'--ko',...
     x,abs(Ga(3,:)),'--bs',...
    'LineWidth',2,'MarkerSize',5)
                     
xlim([0 nx+100])
ylim([0 max(abs(Ga(end,:)))])

h=legend({'$\alpha=10^{-4}$',...
'$\alpha=1$','$\alpha=10^{+4}$',...
},'Interpreter','latex');
xlabel('k')
ylabel('Eigenvalues')

set(h,'FontSize',10,'fontWeight','bold');

set(gca,'FontSize',10,'fontWeight','bold');

set(findall(gcf,'type','text'),'FontSize',10,'fontWeight','bold')

hold off

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Plot for a fixed alpha')
hold on
x=1:n*P*nt;
nx =length(x);
%For beta
plot(x,abs(Gb(1,:)),'--g*', x,abs(Gb(2,:)),'--k*',...
     x,abs(Gb(3,:)),'--bo',...
     x,abs(Gb(4,:)),'--cs',...
     x,abs(Gb(5,:)),'--rs','LineWidth',2,'MarkerSize',5)
 
                     
xlim([0 nx+100])
%ylim([0 max(abs(Gb(5,:)))])
h2=legend({'$\beta=10^{-8}$','$\beta=10^{-4}$',...
'$\beta=1$','$\beta=10^{+4}$','$\beta=10^{+8}$',...
},'Interpreter','latex');

xlabel('k')
ylabel('Eigenvalues')

set(h2,'FontSize',10,'fontWeight','bold');

set(gca,'FontSize',10,'fontWeight','bold');

set(findall(gcf,'type','text'),'FontSize',10,'fontWeight','bold')

hold off



