% A runfile for carrying out a series of 2d experiments.
% Each serie corresponds to one varying parameters:
% exptype = 'nt,T,beta,alpha,tol,nu1,ns,K'
% expindex can be any vector with values from 1 to the maximal amount of
% experiments in each serie

function test_expseries_2d(exptype,expindex)

maxNumCompThreads(1);

exptype
expindex = reshape(expindex, 1, [])

%% Experiment for nt
if (~isempty(strfind(exptype, 'nt')))
global_ttimes = zeros(0,1);
global_resids = zeros(0,3);
global_errs = zeros(0,3);
global_ttranks = zeros(0,1);
global_mem = zeros(0,1);
global_errobs = zeros(0,1);
global_moments = zeros(0,2); % mean, std
global_iters = zeros(0,1);
[N,T,beta,alpha,tol,nu1]=defparams;
cnt = 1;
Ns = (5:12)';
for iexp=expindex
    N = Ns(iexp)
    tol = tol*0.1;
    oneexp_2d
    sol1_ex = sol1;
    sol2_ex = sol2;
    sol3_ex = sol3;
    tol = tol*10;
    oneexp_2d
    err1 = norm(sol1-sol1_ex)/norm(sol1_ex)
    err2 = norm(sol2-sol2_ex)/norm(sol2_ex)
    err3 = norm(sol3-sol3_ex)/norm(sol3_ex)
    
    global_ttimes(cnt,1) = ttimes;
    global_resids(cnt,:) = [resid1, resid2, resid3];
    global_errs(cnt,:) = [err1, err2, err3];
    global_ttranks(cnt,1) = max(sol.r);
    global_mem(cnt,1) = mem(sol);
    global_errobs(cnt,1) = err_obs;
    global_moments(cnt,:) = [sol1_mean, sol1_std];
    global_iters(cnt,1) = spat_iters;
    cnt=cnt+1;
end;

dat = [Ns(expindex), global_ttimes, max(global_resids,[],2), global_errs, global_ttranks, global_mem, global_errobs, global_moments, global_iters];
save(sprintf('stokescontrol_nt_%d_%d-%d.dat', np, min(expindex), max(expindex)), '-ascii', 'dat');
fprintf('Data for TikZ is saved to stokescontrol_nt_%d_%d-%d.dat\n', np, min(expindex), max(expindex));
end;

%% Experiment for T
if (~isempty(strfind(exptype, 'T')))
global_ttimes = zeros(0,1);
global_resids = zeros(0,3);
global_errs = zeros(0,3);
global_ttranks = zeros(0,1);
global_mem = zeros(0,1);
global_errobs = zeros(0,1);
global_moments = zeros(0,2); % mean, std
global_iters = zeros(0,1);
[N,T,beta,alpha,tol,nu1]=defparams;
cnt = 1;
Ts = [1e-2,1e-1,1e0,1e1,1e2]';
for iexp=expindex
    T = Ts(iexp)
    tol = tol*0.1;
    oneexp_2d
    sol1_ex = sol1;
    sol2_ex = sol2;
    sol3_ex = sol3;
    tol = tol*10;
    oneexp_2d
    err1 = norm(sol1-sol1_ex)/norm(sol1_ex)
    err2 = norm(sol2-sol2_ex)/norm(sol2_ex)
    err3 = norm(sol3-sol3_ex)/norm(sol3_ex)
    
    global_ttimes(cnt,1) = ttimes;
    global_resids(cnt,:) = [resid1, resid2, resid3];
    global_errs(cnt,:) = [err1, err2, err3];
    global_ttranks(cnt,1) = max(sol.r);
    global_mem(cnt,1) = mem(sol);
    global_errobs(cnt,1) = err_obs;
    global_moments(cnt,:) = [sol1_mean, sol1_std];
    global_iters(cnt,1) = spat_iters;
    cnt=cnt+1;
end;

dat = [Ts(expindex), global_ttimes, max(global_resids,[],2), global_errs, global_ttranks, global_mem, global_errobs, global_moments, global_iters];
save(sprintf('stokescontrol_T_%d_%d-%d.dat', np, min(expindex), max(expindex)), '-ascii', 'dat');
fprintf('Data for TikZ is saved to stokescontrol_T_%d_%d-%d.dat\n', np, min(expindex), max(expindex));
end;


%% Experiment for beta
if (~isempty(strfind(exptype, 'beta')))
global_ttimes = zeros(0,1);
global_resids = zeros(0,3);
global_errs = zeros(0,3);
global_ttranks = zeros(0,1);
global_mem = zeros(0,1);
global_errobs = zeros(0,1);
global_moments = zeros(0,2); % mean, std
global_iters = zeros(0,1);
[N,T,beta,alpha,tol,nu1]=defparams;
cnt = 1;
betas=[1e3,1e2,1e1,1e0,1e-2,1e-4,1e-6,1e-8,1e-10]';
for iexp=expindex
    beta = betas(iexp)
    tol = tol*0.1;
    oneexp_2d
    sol1_ex = sol1;
    sol2_ex = sol2;
    sol3_ex = sol3;
    tol = tol*10;
    oneexp_2d
    err1 = norm(sol1-sol1_ex)/norm(sol1_ex)
    err2 = norm(sol2-sol2_ex)/norm(sol2_ex)
    err3 = norm(sol3-sol3_ex)/norm(sol3_ex)
    
    global_ttimes(cnt,1) = ttimes;
    global_resids(cnt,:) = [resid1, resid2, resid3];
    global_errs(cnt,:) = [err1, err2, err3];
    global_ttranks(cnt,1) = max(sol{1}.r);
    global_mem(cnt,1) = mem(sol{1});
    global_errobs(cnt,1) = err_obs;
    global_moments(cnt,:) = [sol1_mean, sol1_std];
    global_iters(cnt,1) = spat_iters;
    cnt=cnt+1;
end;

dat = [betas(expindex), global_ttimes, max(global_resids,[],2), global_errs, global_ttranks, global_mem, global_errobs, global_moments, global_iters];
save(sprintf('stokescontrol_beta_%d_%d-%d.dat', np, min(expindex), max(expindex)), '-ascii', 'dat');
fprintf('Data for TikZ is saved to stokescontrol_beta_%d_%d-%d.dat\n', np, min(expindex), max(expindex));
end;

%% Experiment for alpha
if (~isempty(strfind(exptype, 'alpha')))
global_ttimes = zeros(0,1);
global_resids = zeros(0,3);
global_errs = zeros(0,3);
global_ttranks = zeros(0,1);
global_mem = zeros(0,1);
global_errobs = zeros(0,1);
global_moments = zeros(0,2); % mean, std
global_iters = zeros(0,1);
[N,T,beta,alpha,tol,nu1]=defparams;

cnt = 1;
alphas=[0,1e-3,1e-2,1e-1,1e0,1e1,1e2,1e3]';
for iexp=expindex
    alpha = alphas(iexp)
    tol = tol*0.1;
    oneexp_2d
    sol1_ex = sol1;
    sol2_ex = sol2;
    sol3_ex = sol3;
    tol = tol*10;
    oneexp_2d
    err1 = norm(sol1-sol1_ex)/norm(sol1_ex)
    err2 = norm(sol2-sol2_ex)/norm(sol2_ex)
    err3 = norm(sol3-sol3_ex)/norm(sol3_ex)
    
    global_ttimes(cnt,1) = ttimes;
    global_resids(cnt,:) = [resid1, resid2, resid3];
    global_errs(cnt,:) = [err1, err2, err3];
    global_ttranks(cnt,1) = max(sol.r);
    global_mem(cnt,1) = mem(sol);
    global_errobs(cnt,1) = err_obs;
    global_moments(cnt,:) = [sol1_mean, sol1_std];
    global_iters(cnt,1) = spat_iters;
    cnt=cnt+1;
end;

dat = [alphas(expindex), global_ttimes, max(global_resids,[],2), global_errs, global_ttranks, global_mem, global_errobs, global_moments, global_iters];
save(sprintf('stokescontrol_alpha_%d_%d-%d.dat', np, min(expindex), max(expindex)), '-ascii', 'dat');
fprintf('Data for TikZ is saved to stokescontrol_alpha_%d_%d-%d.dat\n', np, min(expindex), max(expindex));
end;

%% Experiment for tol
if (~isempty(strfind(exptype, 'tol')))
global_ttimes = zeros(0,1);
global_resids = zeros(0,3);
global_errs = zeros(0,3);
global_ttranks = zeros(0,1);
global_mem = zeros(0,1);
global_errobs = zeros(0,1);
global_moments = zeros(0,2); % mean, std
global_iters = zeros(0,1);
[N,T,beta,alpha,tol,nu1]=defparams;
cnt = 1;
sol1_ex = [];
tols=[1e-8,1e-7,1e-6,1e-5,1e-4]';
for iexp=expindex
    tol = tols(iexp)
    oneexp_2d
    if (isempty(sol1_ex))
        sol1_ex = sol1;
        sol2_ex = sol2;
        sol3_ex = sol3;
    end;
    err1 = norm(sol1-sol1_ex)/norm(sol1_ex)
    err2 = norm(sol2-sol2_ex)/norm(sol2_ex)
    err3 = norm(sol3-sol3_ex)/norm(sol3_ex)    
    
    global_ttimes(cnt,1) = ttimes;
    global_resids(cnt,:) = [resid1, resid2, resid3];
    global_errs(cnt,:) = [err1, err2, err3];
    global_ttranks(cnt,1) = max(sol.r);
    global_mem(cnt,1) = mem(sol);
    global_errobs(cnt,1) = err_obs;
    global_moments(cnt,:) = [sol1_mean, sol1_std];
    global_iters(cnt,1) = spat_iters;
    cnt=cnt+1;
end;

dat = [tols(expindex), global_ttimes, max(global_resids,[],2), global_errs, global_ttranks, global_mem, global_errobs, global_moments, global_iters];
save(sprintf('stokescontrol_tol_%d_%d-%d.dat', np, min(expindex), max(expindex)), '-ascii', 'dat');
fprintf('Data for TikZ is saved to stokescontrol_tol_%d_%d-%d.dat\n', np, min(expindex), max(expindex));
end;

%% Experiment for nu1
if (~isempty(strfind(exptype, 'nu1')))
global_ttimes = zeros(0,1);
global_resids = zeros(0,3);
global_errs = zeros(0,3);
global_ttranks = zeros(0,1);
global_mem = zeros(0,1);
global_errobs = zeros(0,1);
global_moments = zeros(0,2); % mean, std
global_iters = zeros(0,1);
[N,T,beta,alpha,tol,nu1]=defparams;
cnt = 1;
nu1s=[0,1e-4,1e-3,1e-2,1e-1,3e-1,7e-1,9e-1]';
for iexp=expindex
    nu1 = nu1s(iexp)
    tol = tol*0.1;
    oneexp_2d
    sol1_ex = sol1;
    sol2_ex = sol2;
    sol3_ex = sol3;
    tol = tol*10;
    oneexp_2d
    err1 = norm(sol1-sol1_ex)/norm(sol1_ex)
    err2 = norm(sol2-sol2_ex)/norm(sol2_ex)
    err3 = norm(sol3-sol3_ex)/norm(sol3_ex)
    
    global_ttimes(cnt,1) = ttimes;
    global_resids(cnt,:) = [resid1, resid2, resid3];
    global_errs(cnt,:) = [err1, err2, err3];
    global_ttranks(cnt,1) = max(sol.r);
    global_mem(cnt,1) = mem(sol);
    global_errobs(cnt,1) = err_obs;
    global_moments(cnt,:) = [sol1_mean, sol1_std];
    global_iters(cnt,1) = spat_iters;
    cnt=cnt+1;
end;

dat = [nu1s(expindex), global_ttimes, max(global_resids,[],2), global_errs, global_ttranks, global_mem, global_errobs, global_moments, global_iters];
save(sprintf('stokescontrol_nu1_%d_%d-%d.dat', np, min(expindex), max(expindex)), '-ascii', 'dat');
fprintf('Data for TikZ is saved to stokescontrol_nu1_%d_%d-%d.dat\n', np, min(expindex), max(expindex));
end;


%% Experiment for ns
if (~isempty(strfind(exptype, 'ns')))
global_ttimes = zeros(0,1);
global_resids = zeros(0,3);
global_errs = zeros(0,3);
global_ttranks = zeros(0,1);
global_mem = zeros(0,1);
global_errobs = zeros(0,1);
global_moments = zeros(0,2); % mean, std
global_iters = zeros(0,1);
[N,T,beta,alpha,tol,nu1]=defparams;
cnt = 1;
nss=[16,32,64,128]';
for iexp=expindex
    ns = nss(iexp)
    tol = tol*0.1;
    oneexp_2d
    sol1_ex = sol1;
    sol2_ex = sol2;
    sol3_ex = sol3;
    tol = tol*10;
    oneexp_2d
    err1 = norm(sol1-sol1_ex)/norm(sol1_ex)
    err2 = norm(sol2-sol2_ex)/norm(sol2_ex)
    err3 = norm(sol3-sol3_ex)/norm(sol3_ex)
    
    global_ttimes(cnt,1) = ttimes;
    global_resids(cnt,:) = [resid1, resid2, resid3];
    global_errs(cnt,:) = [err1, err2, err3];
    global_ttranks(cnt,1) = max(sol{1}.r);
    global_mem(cnt,1) = mem(sol{1});
    global_errobs(cnt,1) = err_obs;
    global_moments(cnt,:) = [sol1_mean, sol1_std];
    global_iters(cnt,1) = spat_iters;
    cnt=cnt+1;
end;

dat = [nss(expindex), global_ttimes, max(global_resids,[],2), global_errs, global_ttranks, global_mem, global_errobs, global_moments, global_iters];
save(sprintf('stokescontrol_ns_%d_%d-%d.dat', np, min(expindex), max(expindex)), '-ascii', 'dat');
fprintf('Data for TikZ is saved to stokescontrol_ns_%d_%d-%d.dat\n', np, min(expindex), max(expindex));
end;


%% Experiment for K
if (~isempty(strfind(exptype, 'K')))
global_ttimes = zeros(0,1);
global_resids = zeros(0,3);
global_errs = zeros(0,3);
global_ttranks = zeros(0,1);
global_mem = zeros(0,1);
global_errobs = zeros(0,1);
global_moments = zeros(0,2); % mean, std
global_iters = zeros(0,1);
[N,T,beta,alpha,tol,nu1]=defparams;
cnt = 1;
Ks=[0,1,2,4,6,8,10]';
for iexp=expindex
    K = Ks(iexp)
    tol = tol*0.1;
    oneexp_2d
    sol1_ex = sol1;
    sol2_ex = sol2;
    sol3_ex = sol3;
    tol = tol*10;
    oneexp_2d
    err1 = norm(sol1-sol1_ex)/norm(sol1_ex)
    err2 = norm(sol2-sol2_ex)/norm(sol2_ex)
    err3 = norm(sol3-sol3_ex)/norm(sol3_ex)
    
    global_ttimes(cnt,1) = ttimes;
    global_resids(cnt,:) = [resid1, resid2, resid3];
    global_errs(cnt,:) = [err1, err2, err3];
    global_ttranks(cnt,1) = max(sol.r);
    global_mem(cnt,1) = mem(sol);
    global_errobs(cnt,1) = err_obs;
    global_moments(cnt,:) = [sol1_mean, sol1_std];
    global_iters(cnt,1) = spat_iters;
    cnt=cnt+1;
end;

dat = [Ks(expindex), global_ttimes, max(global_resids,[],2), global_errs, global_ttranks, global_mem, global_errobs, global_moments, global_iters];
save(sprintf('stokescontrol_K_%d_%d-%d.dat', np, min(expindex), max(expindex)), '-ascii', 'dat');
fprintf('Data for TikZ is saved to stokescontrol_K_%d_%d-%d.dat\n', np, min(expindex), max(expindex));
end;

end

function [N,T,beta,alpha,tol,nu1]=defparams
% Default parameters
% Time dimension
N = 10;
% Time interval
T = 1;
% Regularization
beta = 1e-6;
alpha = 1;
% Tolerance 
tol = 1e-6;
% Stochastics
nu1 = 1e-1;
end
